# SPDX-License-Identifier: GNU GPL v3
"""
A class for building an image filter search space for graph generation.
"""

import random
import numpy as np
from dataclasses import dataclass
from ..imaging.base_image import BaseImage
from ..utils.sgt_utils import AbortException


class FilterSearchSpace:
    """
    Class for building a discrete search space of image filters. This search space is huge and irregular
    (over 11k Trillion candidates) and does not have the structure of (Markov Decision Process) MDP states. For example,
    if the current state S1 is a combination of image filter configurations, then the decision to select the
    next or future states S2, S3, ... does not depend on the previous state S1 (No Markov Property).

    No Markov Property: picking the next candidate (a combination of image filter configurations) does not depend on the
    previous candidate. No future candidate/state/action is prohibited (or determined) by the
    current candidate/state/action.

    Markov Property: current candidate determines the next future candidates.

    For this reason, we use Genetic Algorithm (GA) to find the best combination of image filter configurations. GA is
    a global optimizer (or global optimization method/algorithm) that finds the best solution in a given search space.
    """

    @dataclass
    class Candidate:
        """A candidate in the search space. It contains a position in the search space and, the Standard Deviation (SD)
        of the pixel values."""
        position: int | None = None
        std_cost: float | None = None

    @dataclass
    class SearchSpace:
        """Discrete search space of image filters; where, each candidate is a combination of image filter
        configurations. We use this template to build 3 search spaces: apply filters, value filters, and brightness
        filters."""
        min_pos: int = 0
        max_pos: int = 0
        candidates: list = None
        ignore_candidates: list= None
        best_candidate = None

    @dataclass
    class FilterCandidate:
        """A filter candidate in the search space. It contains
        a position in the search space (which encodes a binary number), the position determines the value range of the
        value search space. It also has the brightness search space, and the cost of applying the filter is calculated by
        evaluating the binary image and finding the number of white pixels in the image. Retrieve the corresponding pixel
        values from the original image and calculate the Standard Deviation (SD) of the pixel values. Finally, it has
        the combination of image filter configurations."""
        position: int | None = None               # 11 bits long (approx. 2k candidates)
        value_range: tuple[int, int] | None = None                  # [min, max] values -- 0bits-20bits
        # value_candidate: "FilterSearchSpace.Candidate" = None
        # brightness_candidate: "FilterSearchSpace.Candidate" = None
        value_space: "FilterSearchSpace.SearchSpace" = None         # approx. 1B candidates
        brightness_space: "FilterSearchSpace.SearchSpace" = None    # approx. 256 candidates
        std_cost: float | None = None
        graph_accuracy: float | None = None     # CNN model prediction accuracy of the generated graph
        img_configs: dict | None = None

    def __init__(self):
        pass

    @classmethod
    def _build_full_search_space(cls, img_obj: BaseImage) -> SearchSpace | None:
        """
        Create a discrete search space where each candidate is a combination of image filter configurations.
        The actual search space has over 118k Trillion candidates. This method is used for debugging purposes -- the
        search space is too large to be used in production.

        :param img_obj: The image object.
        :return: The search space.
        """
        if img_obj is None:
            return None

        # Set ranges for each parameter (Discrete action space)
        threshold_types = [0, 1, 2]  # global / adaptive / OTSU
        global_thresh_range = list(range(1, 256))  # 1–255
        adaptive_local_range = list(range(1, 100, 2))  # 1–99 (odd)
        brightness_levels = list(range(-100, 101))  # -100–100
        contrast_levels = list(range(-100, 101))  # -100–100
        gamma_range = np.arange(0.01, 5.01, 0.01)  # 0.01–5.0
        blurring_window_sizes = list(range(1, 8, 2))  # 1, 3, 7 (odd)
        filter_window_sizes = list(range(1, 101))  # 1–100

        # Initialize search space
        pos = 0
        init_configs = img_obj.configs.copy()
        search_space = FilterSearchSpace.SearchSpace(candidates=[], ignore_candidates=[])

        for tt in threshold_types:
            global_range = global_thresh_range if tt == 0 else [128]
            adaptive_range = adaptive_local_range if tt == 1 else [11]
            for global_thresh in global_range:
                for adaptive_thresh in adaptive_range:
                    for brightness in brightness_levels:
                        for contrast in contrast_levels:
                            for gamma_val in gamma_range:
                                for blur_size in blurring_window_sizes:
                                    for filter_size in filter_window_sizes:
                                        init_configs["threshold_type"]["value"] = tt
                                        init_configs["global_threshold_value"]["value"] = global_thresh
                                        init_configs["adaptive_local_threshold_value"]["value"] = adaptive_thresh
                                        init_configs["brightness_level"]["value"] = brightness
                                        init_configs["contrast_level"]["value"] = contrast
                                        init_configs["apply_gamma"]["dataValue"] = gamma_val
                                        init_configs["apply_autolevel"]["dataValue"] = blur_size
                                        init_configs["apply_gaussian_blur"]["dataValue"] = blur_size
                                        init_configs["apply_lowpass_filter"]["dataValue"] = filter_size
                                        init_configs["apply_laplacian_gradient"]["dataValue"] = blur_size
                                        init_configs["apply_sobel_gradient"]["dataValue"] = blur_size
                                        for apply_dark_fg in [0, 1]:
                                            for apply_gamma in [0, 1]:
                                                for apply_auto_lvl in [0, 1]:
                                                    for apply_laplacian in [0, 1]:
                                                        for apply_gaussian in [0, 1]:
                                                            for apply_lowpass in [0, 1]:
                                                                for apply_sobel in [0, 1]:
                                                                    for apply_median in [0, 1]:
                                                                        for apply_scharr in [0, 1]:
                                                                            init_configs["apply_dark_foreground"][
                                                                                "value"] = apply_dark_fg
                                                                            init_configs["apply_gamma"][
                                                                                "value"] = apply_gamma
                                                                            init_configs["apply_autolevel"][
                                                                                "value"] = apply_auto_lvl
                                                                            init_configs["apply_laplacian_gradient"][
                                                                                "value"] = apply_laplacian
                                                                            init_configs["apply_gaussian_blur"][
                                                                                "value"] = apply_gaussian
                                                                            init_configs["apply_lowpass_filter"][
                                                                                "value"] = apply_lowpass
                                                                            init_configs["apply_sobel_gradient"][
                                                                                "value"] = apply_sobel
                                                                            init_configs["apply_median_filter"][
                                                                                "value"] = apply_median
                                                                            init_configs["apply_scharr_gradient"][
                                                                                "value"] = apply_scharr
                                                                            # candidate = FilterSearchSpace.Candidate(
                                                                            #     position=pos,
                                                                            #     std_cost=None,
                                                                            #     img_configs=init_configs.copy()
                                                                            # )
                                                                            # search_space.candidates.append(candidate)
                                                                            print(
                                                                                f"Candidate {pos} added to search space.")
                                                                            pos += 1
        return search_space

    @staticmethod
    def get_initial_candidate(search_space: "FilterSearchSpace.SearchSpace"):
        """
        Get the initial candidate.

        :param search_space: The search space.
        """
        idx = random.randint(0, len(search_space.candidates) - 1)
        if search_space.best_candidate is None:
            init_candidate = search_space.candidates[idx]
        elif search_space.best_candidate.position in search_space.ignore_candidates:
            init_candidate = search_space.candidates[idx]
        else:
            init_candidate = search_space.best_candidate
        init_candidate.std_cost = np.inf
        return init_candidate

    @staticmethod
    def decode_candidate_position(encoded_pos: int, img_configs: dict) -> None|dict:
        """
        Decode the position of a candidate in the search space into a dictionary of image filter configurations.

        :param encoded_pos: The position of the candidate in the search space.
        :param img_configs: The dictionary of image filter configurations.
        :return: The dictionary of image filter configurations.
        """
        if encoded_pos is None or img_configs is None:
            return None

            # Step 1: Convert integer to 11-bit binary string
        bitstring = format(encoded_pos, "011b")  # always 11 bits

        # Step 2: Extract threshold type (first 2 bits)
        threshold_bits = bitstring[:2]
        threshold_type = int(threshold_bits, 2)

        # Step 3: Extract 9 filter bits (order consistent with the encoding)
        filter_bits = bitstring[2:]
        filters = [int(b) for b in filter_bits]

        # Step 4: Map to variable names
        img_configs["threshold_type"]["value"] = threshold_type
        img_configs["apply_dark_foreground"]["value"] = filters[0]
        img_configs["apply_gamma"]["value"] = filters[1]
        img_configs["apply_autolevel"]["value"] = filters[2]
        img_configs["apply_laplacian_gradient"]["value"] = filters[3]
        img_configs["apply_gaussian_blur"]["value"] = filters[4]
        img_configs["apply_lowpass_filter"]["value"] = filters[5]
        img_configs["apply_sobel_gradient"]["value"] = filters[6]
        img_configs["apply_median_filter"]["value"] = filters[7]
        img_configs["apply_scharr_gradient"]["value"] = filters[8]

        """
        # Step 5: Build value_bits for extra parameter ranges
        value_bits = ""
        if img_configs["apply_gamma"]["value"] == 1:
            value_bits += format(500, "09b")  # gamma parameter
        if img_configs["apply_autolevel"]["value"] == 1:
            value_bits += format(1, "03b")  # autolevel param
        if img_configs["apply_gaussian_blur"]["value"] == 1:
            value_bits += format(1, "03b")  # blur strength
        if img_configs["apply_lowpass_filter"]["value"] == 1:
            value_bits += format(3, "07b")  # filter kernel

        # Step 6: Convert extra bits to integer (if any)
        max_val = int(value_bits, 2) if value_bits else 0
        """
        return img_configs

    @staticmethod
    def decode_filter_values(img_configs: dict, value_candidate: "FilterSearchSpace.Candidate"=None, bright_candidate: "FilterSearchSpace.Candidate"=None) -> dict|None:
        """
        Decode the image filter configurations of a candidate into a dictionary.

        :param img_configs: The dictionary of image filter configurations.
        :param value_candidate: The filter-value candidate for updating image filter value configurations.
        :param bright_candidate: The brightness-value candidate for updating image brightness/contrast configurations.
        :return: The dictionary of image filter configurations.
        """
        if img_configs is None:
            return None

        if value_candidate is not None:
            encoded_val_pos = value_candidate.position
            if encoded_val_pos is None:
                return None
            bit_str = format(encoded_val_pos, "030b")
            threshold_val = int(bit_str[:8], 2)
            gamma_val = int(bit_str[8:17], 2)
            autolevel_val = int(bit_str[17:19], 2)
            gaussian_val = int(bit_str[19:21], 2)
            laplacian_val = int(bit_str[21:23], 2)
            lowpass_val = int(bit_str[23:30], 2)
            blur_window_size = [1, 3, 5, 7]

            if img_configs["threshold_type"]["value"] == 0:
                img_configs["global_threshold_value"]["value"] = threshold_val

            if img_configs["threshold_type"]["value"] == 1:
                # Should be an Odd number
                threshold_val = threshold_val+1 if threshold_val%2==0 else threshold_val
                img_configs["adaptive_local_threshold_value"]["value"] = threshold_val

            if img_configs["apply_gamma"]["value"] == 1:
                img_configs["apply_gamma"]["dataValue"] = round(gamma_val / 100.0, 2) if gamma_val > 0 else 0.01

            if img_configs["apply_lowpass_filter"]["value"] == 1:
                img_configs["apply_lowpass_filter"]["dataValue"] = lowpass_val

            if img_configs["apply_autolevel"]["value"] == 1:
                img_configs["apply_autolevel"]["dataValue"] = blur_window_size[autolevel_val]

            if img_configs["apply_gaussian_blur"]["value"] == 1:
                img_configs["apply_gaussian_blur"]["dataValue"] = blur_window_size[gaussian_val]

            if img_configs["apply_laplacian_gradient"]["value"] == 1:
                img_configs["apply_laplacian_gradient"]["dataValue"] = blur_window_size[laplacian_val]

            if img_configs["apply_sobel_gradient"]["value"] == 1:
                # To Be Updated (we need to keep bit_str less than 30 bits)
                img_configs["apply_sobel_gradient"]["dataValue"] = blur_window_size[laplacian_val] # re-use laplacian gradient

        if bright_candidate is not None:
            encoded_brightness_pos = bright_candidate.position
            if encoded_brightness_pos is None:
                return None
            bit_str = format(encoded_brightness_pos, "016b")
            is_brightness_neg = bit_str[0]
            brightness_val = int(bit_str[1:8], 2)
            if is_brightness_neg == "1":
                brightness_val = -brightness_val

            is_contrast_neg = bit_str[8]
            contrast_val = int(bit_str[9:16], 2)
            if is_contrast_neg == "1":
                contrast_val = -contrast_val

            img_configs["contrast_level"]["value"] = contrast_val
            img_configs["brightness_level"]["value"] = brightness_val
        return img_configs

    @staticmethod
    def build_search_space(img_obj: BaseImage, initial_pop: int = 256) -> SearchSpace | None:
        """
        Create a discrete search space where each candidate is a combination of image filter configurations.
        Encodes a combination of image filter configurations as a binary number, then this number is converted into an
        integer position in the search space.

        :param img_obj: The image object.
        :param initial_pop: The total population size for Genetic Algorithm search space. Default is 256.
        :return: The search space.
        """

        """
        def encode_filter_combination(
                threshold_type=1,  # 0, 1, or 2 → needs 2 bits
                apply_dark_foreground=0,
                apply_gamma=1,
                apply_auto_level=0,
                apply_laplacian_gradient=0,
                apply_gaussian_blur=0,
                apply_lowpass_filter=0,
                apply_sobel_gradient=0,
                apply_median_filter=0,
                apply_scharr_gradient=0,
        )-> tuple[str, int]:
            # Encode 10 image filter configurations as an 11-bit binary string (2 bits for the threshold type,
            # 9 bits for filters). The total number of filter combinations is 2^11 = 2048.
            # :returns: Both the binary string and integer representation.

            # --- Step 1: Encode threshold_type into 2 bits ---
            if threshold_type not in [0, 1, 2]:
                raise ValueError("threshold_type must be 0, 1, or 2")
            threshold_bits = format(threshold_type, "02b")  # 2-bit binary

            # --- Step 2: Encode 9 filters into 1 bit each ---
            filters = [
                apply_dark_foreground,
                apply_gamma,
                apply_auto_level,
                apply_laplacian_gradient,
                apply_gaussian_blur,
                apply_lowpass_filter,
                apply_sobel_gradient,
                apply_median_filter,
                apply_scharr_gradient,
            ]

            filter_bits = "".join(str(int(f)) for f in filters)

            # --- Step 3: Concatenate ---
            bitstring = threshold_bits + filter_bits

            # --- Step 4: Convert to integer ---
            bit_int = int(bitstring, 2)
            return bitstring, bit_int
        """

        if img_obj is None:
            return None

        # Parameters
        apply_pop = 2**11
        val_range = (2**22, 2**30)  # minimum, maximum value range for search space
        bri_range = (0, 2**16)
        # print(f"Apply candidates: {apply_pop}, Value candidates: {val_range}, Brightness candidates: {bri_range}")

        # Initialize search space
        init_configs = img_obj.configs.copy()
        search_space = FilterSearchSpace.SearchSpace(candidates=[], ignore_candidates=[], min_pos=0, max_pos=apply_pop-1)
        for pos in range(apply_pop):
            img_configs = FilterSearchSpace.decode_candidate_position(pos, init_configs)
            if img_configs is not None:
                # Create an empty candidate
                val_pop = [FilterSearchSpace.Candidate(position=random.randrange(val_range[0], val_range[1]), std_cost=None) for _ in range(initial_pop)]
                b_pop = [FilterSearchSpace.Candidate(position=random.randrange(bri_range[0], bri_range[1]), std_cost=None) for _ in range(initial_pop)]

                filter_candidate = FilterSearchSpace.FilterCandidate(
                    position=pos,
                    value_range=(0, initial_pop),
                    value_space=FilterSearchSpace.SearchSpace(candidates=val_pop, ignore_candidates=[], min_pos=val_range[0], max_pos=val_range[1]),
                    brightness_space=FilterSearchSpace.SearchSpace(candidates=b_pop, ignore_candidates=[], min_pos=bri_range[0], max_pos=bri_range[1]),
                    std_cost=None,
                    graph_accuracy=None,
                    img_configs=img_configs.copy(),
                )
                search_space.candidates.append(filter_candidate)
                if pos == 640:
                    # default candidate
                    search_space.best_candidate = filter_candidate
                    search_space.best_candidate.value_space.best_candidate = FilterSearchSpace.Candidate(position=2**22, std_cost=None)
                    search_space.best_candidate.brightness_space.best_candidate = FilterSearchSpace.Candidate(position=0, std_cost=None)
        return search_space

    @staticmethod
    def cost_function(new_img_configs: dict, img_obj: BaseImage) -> float:
        """Calculate and apply the cost of a candidate. Given the image filter configurations, apply them to get a
        binary image and find the number of white pixels in the image. Retrieve the corresponding pixel values from the
        original image and calculate the Standard Deviation (SD) of the pixel values.

        :param new_img_configs: The dictionary of image filter configurations.
        :param img_obj: The image object.
        """

        if img_obj is None:
            return np.inf

        if new_img_configs is None:
            return np.inf

        # Copy image filter configurations to the image object
        img_obj.configs = new_img_configs.copy()
        # Reset image filters
        img_obj.img_mod, img_obj.img_bin = None, None
        # Apply image filters
        img_data = img_obj.img_2d.copy()
        img_obj.img_mod = img_obj.process_img(image=img_data)
        img_mod = img_obj.img_mod.copy()
        img_obj.img_bin = img_obj.binarize_img(img_mod)
        img_obj.img_mod = img_mod
        # Compute SD as cost
        eval_std, eval_hist = img_obj.evaluate_img_binary()
        return eval_std

    @staticmethod
    def evaluate_candidate(search_space: "FilterSearchSpace.SearchSpace",
                           candidate: "FilterSearchSpace.Candidate") -> None:
        """
        Evaluate a candidate in the search space, check if it is better than the best candidate.

        :param search_space: The search space.
        :param candidate: The candidate to evaluate.
        """
        if candidate.std_cost is None:
            return

        if search_space.best_candidate is None:
            search_space.best_candidate = FilterSearchSpace.Candidate(
                position=candidate.position,
                std_cost=candidate.std_cost
            )

        if candidate.position in search_space.ignore_candidates:
            return

        elif candidate.std_cost < search_space.best_candidate.std_cost:
            search_space.best_candidate = FilterSearchSpace.Candidate(
                position=candidate.position,
                std_cost=candidate.std_cost,
            )


def sgt_genetic_algorithm(s_space: FilterSearchSpace.SearchSpace, img_obj: BaseImage, generations: int = 4, pop_size: int = 8, gamma: float = 1.0, mu: float = 0.9, sigma: float = 0.9) -> dict|None:
    """
    Executes the genetic algorithm to find the best candidate from a huge search space.

    :param s_space: Search space object.
    :param img_obj: BaseImage object which contains the image itself and the image configurations.
    :param generations: Number of family generations to run the algorithm for.
    :param pop_size: Initial size of the population.
    :param gamma: Crossover probability.
    :param mu: Mutation probability.
    :param sigma: Standard deviation of the Gaussian mutation.

    :return: A dictionary containing the best candidate's image configuration settings.
    """

    def _select_parents():
        """Select parents for crossover."""

        # Select a random parent population (1/3 of the population)
        q = np.random.permutation(pop_size)
        parent_pop = []
        for i in range(pop_size//3):
            parent_pop.append(s_space.candidates[q[i]])
        return parent_pop

    def _crossover(parent_1, parent_2):
        """Cross over two parents to generate two children."""
        if isinstance(parent_1, FilterSearchSpace.Candidate):
            alpha = random.uniform(0, gamma)
            child_1 = FilterSearchSpace.Candidate()
            child_2 = FilterSearchSpace.Candidate()
            # Apply crossover and ensure positions are within bounds
            child_1.position = int(max(s_space.min_pos, min(parent_1.position * alpha + parent_2.position * (1 - alpha), s_space.max_pos)))
            child_2.position = int(max(s_space.min_pos, min(parent_2.position * alpha + parent_1.position * (1 - alpha), s_space.max_pos)))
            return child_1, child_2
        else:
            return parent_1, parent_2

    def _mutate(x):
        """Mutate an individual x to generate a new individual y."""
        if isinstance(x, FilterSearchSpace.Candidate):
            y = FilterSearchSpace.Candidate()
            # Apply Gaussian mutation with mean mu and standard deviation sigma
            mutation_value = np.random.normal(mu, sigma)
            # Mutate the position and ensure it stays within bounds
            y.position = int(np.clip(x.position + mutation_value, s_space.min_pos, s_space.max_pos))
            return y
        else:
            return x

    if s_space is None:
        raise AbortException("Search space cannot be None")

    best_sol = FilterSearchSpace.get_initial_candidate(s_space)
    best_configs = img_obj.configs.copy()
    for _ in range(generations):
        best_individual = None
        temp_configs = None

        # 1. Compute fitness for each individual in the population/search space
        for individual in s_space.candidates:
            if isinstance(individual, FilterSearchSpace.Candidate):
                if s_space.max_pos >= 2**30:
                    new_configs = FilterSearchSpace.decode_filter_values(img_obj.configs.copy(), value_candidate=individual)
                else:
                    new_configs = FilterSearchSpace.decode_filter_values(img_obj.configs.copy(), bright_candidate=individual)
                individual.std_cost = FilterSearchSpace.cost_function(new_configs, img_obj)
                if best_individual is None or best_individual.std_cost is None or individual.std_cost < best_individual.std_cost:
                    best_individual = individual
                    temp_configs = new_configs.copy()

        # 1.1. Update the current best candidate
        if best_individual is None:
            break

        # 1.2. Check if fitness is valid
        if best_individual.std_cost is None:
            break

        # 1.3. Update the current best candidate
        if best_individual is not None and best_individual.std_cost < best_sol.std_cost:
            best_sol = best_individual
            best_configs = temp_configs.copy()

        # 2. Select parents
        parents = _select_parents()

        # 3. Create offspring through crossover and mutation
        new_population = []
        for _ in range(pop_size // 2):
            p_1, p_2 = np.random.choice(parents, size=2, replace=False)

            # 3.1. Crossover parents to generate two children
            c_1, c_2 = _crossover(p_1, p_2)

            # 3.1. Mutate children to generate new candidates
            x_1 = _mutate(c_1)
            x_2 = _mutate(c_2)

            # 3.3. Add children to the new population
            new_population.append(x_1) if x_1.position not in s_space.ignore_candidates else None
            new_population.append(x_2) if x_2.position not in s_space.ignore_candidates else None
        # 4. Apply replacement/elitism if desired
        s_space.candidates = new_population
    s_space.best_candidate = best_sol
    return best_configs



def sgt_hill_climbing_algorithm(s_space: FilterSearchSpace.SearchSpace, img_obj: BaseImage, max_iters: int = 5, step_size: int = 1) -> None:
    """
    Executes the hill climbing algorithm to find the best candidate from a small search space.

    :param s_space: Search space object.
    :param img_obj: BaseImage object which contains the image itself and the image configurations.
    :param max_iters: Maximum number of iterations to run the algorithm for.
    :param step_size: Step size to move the current candidate.

    :return: None
    """

    def _generate_neighbors():
        """Generate neighbors by slightly modifying the current candidate."""
        lst_neighbor = []
        for i in range(5):
            center_pos = best_sol.position
            left_pos = max(s_space.min_pos, center_pos - step_size)
            right_pos = min(s_space.max_pos, center_pos + step_size)
            if isinstance(best_sol, (FilterSearchSpace.Candidate, FilterSearchSpace.FilterCandidate)):
                for item in s_space.candidates:
                    if (item.position in (left_pos, center_pos, right_pos)) and (item.position not in s_space.ignore_candidates):
                        lst_neighbor.append(item)
        return lst_neighbor

    if s_space is None or img_obj is None:
        raise AbortException("Search space or ImageObject cannot be None")

    # 1. Initialize the current best candidate
    init_sol = FilterSearchSpace.get_initial_candidate(s_space)
    if isinstance(s_space.best_candidate, FilterSearchSpace.FilterCandidate):
        best_sol = FilterSearchSpace.FilterCandidate(
            position=init_sol.position,
            value_range=init_sol.value_range,
            value_space=init_sol.value_space,
            brightness_space=init_sol.brightness_space,
            std_cost=np.inf,
            graph_accuracy=0,
            img_configs=init_sol.img_configs,
        )
    else:
        best_sol = FilterSearchSpace.Candidate(position=init_sol.position, std_cost=np.inf)

    # 2. Run the hill climbing algorithm
    for _ in range(max_iters):
        # Get neighbors to the current best candidate
        neighbors = _generate_neighbors()
        best_neighbor = None

        # Find the best neighbor among the neighbors
        for neighbor in neighbors:
            if isinstance(neighbor, FilterSearchSpace.FilterCandidate):
                FilterSearchSpace.decode_candidate_position(neighbor.position, neighbor.img_configs)
                val_sol = neighbor.value_space.best_candidate
                bri_sol = neighbor.brightness_space.best_candidate
                FilterSearchSpace.decode_filter_values(neighbor.img_configs, val_sol, bri_sol)
                neighbor.std_cost = FilterSearchSpace.cost_function(neighbor.img_configs, img_obj)

                if best_neighbor is None or neighbor.std_cost < best_neighbor.std_cost:
                    best_neighbor = neighbor
            elif isinstance(neighbor, FilterSearchSpace.Candidate):
                 new_configs = FilterSearchSpace.decode_filter_values(img_obj.configs.copy(), bright_candidate=neighbor)
                 neighbor.std_cost = FilterSearchSpace.cost_function(new_configs, img_obj)

                 if best_neighbor is None or neighbor.std_cost < best_neighbor.std_cost:
                     best_neighbor = neighbor

        # Update the current best candidate
        if best_neighbor is None:
            break

        if best_neighbor.std_cost is None:
            break

        if best_neighbor is not None and best_neighbor.std_cost < best_sol.std_cost:
            best_sol = best_neighbor
        else:
            # No improvement found, reached a local optimum
            break

    # 3. Update the current best candidate
    s_space.best_candidate = best_sol
    return None



# SPDX-License-Identifier: GNU GPL v3

"""
Base controller class for StructuralGT.
"""

import os
import logging
from PySide6.QtCore import Signal, QObject

from ..utils.sgt_utils import verify_path
from ..imaging.image_processor import ImageProcessor, ALLOWED_IMG_EXTENSIONS
from ..compute.graph_analyzer import GraphAnalyzer

class BaseController(QObject):

    showAlertSignal = Signal(str, str)

    def __init__(self, config_file: str = ""):
        super().__init__()
        # Create graph objects
        self._config_file = config_file
        self._sgt_objs = {}
        self._selected_sgt_obj_index = 0
        self._allow_auto_scale = True

    @property
    def sgt_objs(self):
        return self._sgt_objs

    def replicate_sgt_configs(self) -> None:
        """Replicate the configurations of the selected SGT object to all other SGT objects."""
        # Update Configs
        current_sgt_obj = self.get_selected_sgt_obj()
        if current_sgt_obj is None:
            return

        keys_list = list(self._sgt_objs.keys())
        key_at_current = keys_list[self._selected_sgt_obj_index]
        shared_gtc_configs = current_sgt_obj.configs
        shared_gte_configs = current_sgt_obj.ntwk_p.graph_obj.configs
        shared_img_configs = current_sgt_obj.ntwk_p.image_obj.configs
        for key in keys_list:
            if key != key_at_current:
                s_obj = self._sgt_objs[key]
                s_obj.configs = shared_gtc_configs
                s_obj.ntwk_p.graph_obj.configs = shared_gte_configs
                for img_obj in s_obj.ntwk_p.selected_images:
                    img_obj.configs = shared_img_configs

    def get_selected_sgt_obj(self) -> GraphAnalyzer | None:
        """Retrieve the SGT object at a specified index."""
        try:
            keys_list = list(self._sgt_objs.keys())
            key_at_index = keys_list[self._selected_sgt_obj_index]
            sgt_obj = self._sgt_objs[key_at_index]
            return sgt_obj
        except IndexError:
            logging.info("No Image Error: Please import/add an image.", extra={'user': 'SGT Logs'})
            # self.showAlertSignal.emit("No Image Error", "No image added! Please import/add an image.")
            return None

    def create_sgt_object(self, file_path: str, out_dir: str = "") -> bool:
        """
        A function that processes a selected image file and creates an analyzer object with default configurations.

        Args:
            file_path: file path to the image or the graph data file.
            out_dir: output directory path

        Returns:
        """
        success, result = verify_path(file_path)
        if success:
            file_path = result
            success, result = verify_path(out_dir)
            out_dir = result if success else ""
        else:
            logging.info(result, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("File Error", result)
            return False

        try:
            # Create an SGT object as a GraphAnalyzer object.
            file_ext = os.path.splitext(file_path)[1].lower()
            img_extensions = tuple(ext[1:] if ext.startswith('*.') else ext for ext in ALLOWED_IMG_EXTENSIONS)
            if file_ext.endswith(img_extensions):
                ntwk_p, file_name = ImageProcessor.from_image_file(file_path, out_folder=out_dir,
                                                              config_file=self._config_file,
                                                              allow_auto_scale=self._allow_auto_scale)
            else:
                ntwk_p, file_name = ImageProcessor.from_graph_file(file_path, out_folder=out_dir,)
            sgt_obj = GraphAnalyzer(ntwk_p)

            # Store the StructuralGT object and sync application
            self._sgt_objs[file_name] = sgt_obj
            return True
        except Exception as err:
            logging.exception("File Error: %s", err, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("File Error", "Error processing image. Try again.")
            return False

    def update_output_dir(self, folder_path: str) -> None:
        """Update the output directory for storing StructuralGT results."""
        success, result = verify_path(folder_path)
        if success:
            folder_path = result
        else:
            try:
                os.makedirs(folder_path, exist_ok=True)
            except Exception as err:
                logging.exception("Folder Creation Error: %s", err, extra={'user': 'SGT Logs'})
                self.showAlertSignal.emit("Folder Creation Error", f"Error creating output folder: {err}")
                return

        # Update for all sgt_objs
        key_list = list(self._sgt_objs.keys())
        for key in key_list:
            sgt_obj = self._sgt_objs[key]
            sgt_obj.ntwk_p.output_dir = folder_path

    def add_single_image(self, img_path: str, out_dir: str = "") -> bool:
        """Verify and validate an image path, use it to create an SGT object and load it in view."""
        is_created = self.create_sgt_object(img_path, out_dir)
        if is_created:
            return True
        return False

    def add_multiple_images(self, img_dir_path: str, out_dir: str = "") -> bool:
        """
        Verify and validate multiple image paths, use each to create an SGT object, then load the last one in view.
        """
        success, result = verify_path(img_dir_path)
        if success:
            img_dir_path = result
        else:
            logging.info(result, extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("Directory Error", result)
            return False

        files = os.listdir(img_dir_path)
        files = sorted(files)
        for a_file in files:
            allowed_extensions = tuple(ext[1:] if ext.startswith('*.') else ext for ext in ALLOWED_IMG_EXTENSIONS)
            if a_file.endswith(allowed_extensions):
                img_path = os.path.join(str(img_dir_path), a_file)
                _ = self.create_sgt_object(img_path, out_dir)

        if len(self._sgt_objs) <= 0:
            logging.info("File Error: Files have to be either .tif .png .jpg .jpeg", extra={'user': 'SGT Logs'})
            self.showAlertSignal.emit("File Error",
                                      "No workable images found! Files have to be either .tif, .png, .jpg or .jpeg")
            return False
        else:
            return True

    def add_graph(self, graph_file: str, out_dir: str = "") -> bool:
        """Verify and validate the graph file path, use it to create an SGT object and load it in view."""
        is_created = self.create_sgt_object(graph_file, out_dir)
        if is_created:
            return True
        return False



# SPDX-License-Identifier: GNU GPL v3

"""
Create a graph skeleton from an image binary
"""

import math
import numpy as np
from scipy import ndimage
from cv2.typing import MatLike
from skimage.morphology import binary_dilation as dilate, binary_closing
from skimage.morphology import disk, skeletonize, remove_small_objects


class GraphSkeleton:
    """A class that is used for estimating the width of edges and compute their weights using binerized 2D/3D images."""

    def __init__(self, img_bin: MatLike, configs: dict = None, is_2d: bool = True, progress_func = None):
        """
        A class that builds a skeleton graph from an image.
        The skeleton will be 3D so that it can be analyzed with OVITO

        :param img_bin: OpenCV image in binary format.
        :param configs: Options and parameters.

        >>> import cv2
        >>> import numpy
        >>> opt_gte = {}
        >>> opt_gte["merge_nearby_nodes"]["value"] = 1
        >>> opt_gte["remove_disconnected_segments"]["value"] = 1
        >>> opt_gte["remove_object_size"]["value"] = 500
        >>> opt_gte["prune_dangling_edges"]["value"] = 1
        >>> dummy_image = 127 * numpy.ones((40, 40), dtype = np.uint8)
        >>> img = cv2.threshold(dummy_image, 127, 255, cv2.THRESH_BINARY)[1]
        >>> graph_skel = GraphSkeleton(img, opt_gte)

        """
        self._img_bin = img_bin
        self._configs = configs
        self._is_2d = is_2d
        self._update_progress = progress_func
        self._skeleton, self._skeleton_3d = None, None
        if self._configs is not None:
            self._build_skeleton()

    @property
    def skeleton(self):
        """Returns the skeleton graph in 2D."""
        return self._skeleton

    @property
    def skeleton_3d(self):
        """Returns the skeleton graph in 3D."""
        return self._skeleton_3d

    def _build_skeleton(self) -> None:
        """
        Creates a graph skeleton of the image.

        :return:
        """

        # rebuilding the binary image as a boolean for skeletonizing
        self._img_bin = np.squeeze(self._img_bin)
        img_bin_int = np.asarray(self._img_bin, dtype=np.uint16)

        # making the initial skeleton image
        temp_skeleton = skeletonize(img_bin_int)

        # Use medial axis with distance transform
        # skeleton, distance = medial_axis(img_bin_int, return_distance=True)
        # Scale thickness by distance (optional)
        # temp_skeleton = skeleton * distance

        # if self.configs["remove_bubbles"]["value"] == 1:
        #   if self._update_progress is not None:
        #       self._update_progress([56, f"Removing bubbles from the skeleton..."])
        #   temp_skeleton = GraphSkeleton.remove_bubbles(temp_skeleton, img_bin_int, mask_elements)

        if self._configs["merge_nearby_nodes"]["value"] == 1:
            if self._update_progress is not None:
                self._update_progress([52, f"Merging nearby nodes in the skeleton..."])
            node_radius_size = 2 # int(self.configs["merge_nearby_nodes"]["items"][0]["value"])
            temp_skeleton = GraphSkeleton.merge_nodes(temp_skeleton, node_radius_size)

        if self._configs["remove_disconnected_segments"]["value"] == 1:
            if self._update_progress is not None:
                self._update_progress([54, f"Removing small disconnected segments from the skeleton..."])
            min_size = int(self._configs["remove_disconnected_segments"]["items"][0]["value"])
            temp_skeleton = remove_small_objects(temp_skeleton, min_size=min_size, connectivity=2)

        if self._configs["prune_dangling_edges"]["value"] == 1:
            if self._update_progress is not None:
                self._update_progress([56, f"Pruning dangling edges from skeleton..."])
            max_iter = 500 # int(self.configs["prune_dangling_edges"]["items"][0]["value"])
            b_points = GraphSkeleton.get_branched_points(temp_skeleton)
            temp_skeleton = GraphSkeleton.prune_edges(temp_skeleton, max_iter, b_points)

        self._skeleton = np.asarray(temp_skeleton, dtype=np.uint16)
        # self.skeleton = self.skeleton.astype(int)
        self._skeleton_3d = np.asarray([self._skeleton]) if self._is_2d else self._skeleton

    def assign_weights(self, edge_pts: MatLike, weight_type: str = None, weight_options: dict = None,
                       pixel_dim: float = 1, rho_dim: float = 1) -> tuple[float, float | None, float]:
        """
        Compute and assign weights to a line edge between 2 nodes.

        :param edge_pts: A list of pts that trace along a graph edge.
        :param weight_type: Basis of computation for the weight (i.e., length, width, resistance, conductance, etc.)
        :param weight_options: weight types to be used in computation of weights.
        :param pixel_dim: Physical size of a single pixel width in nanometers.
        :param rho_dim: The resistivity value of the material.
        :return: Width pixel count of edge, computed weight.
        """

        # Initialize parameters: Idea copied from 'sknw' library
        pix_length = np.linalg.norm(edge_pts[1:] - edge_pts[:-1], axis=1).sum()
        epsilon = 0.001             # to avoid division by zero
        pix_length += epsilon

        if len(edge_pts) < 2:
            # check to see if ge is an empty or unity list, if so, set pixel counts to 0
            # Assume only 1/2 pixel exists between edge points
            pix_width = 0.5
            pix_angle = None
        else:
            # if ge exists, find the midpoint of the trace, and orthogonal unit vector
            pix_width, pix_angle = self._estimate_edge_width(edge_pts)
            pix_width += 0.5  # (normalization) to make it larger than empty widths

        if weight_type is None:
            wt = pix_width / 10
        elif weight_options.get(weight_type) == weight_options.get('DIA'):
            wt = pix_width * pixel_dim
        elif weight_options.get(weight_type) == weight_options.get('AREA'):
            wt = math.pi * (pix_width * pixel_dim * 0.5) ** 2
        elif weight_options.get(weight_type) == weight_options.get('LEN') or weight_options.get(weight_type) == weight_options.get('INV_LEN'):
            wt = pix_length * pixel_dim
            if weight_options.get(weight_type) == weight_options.get('INV_LEN'):
                wt = wt + epsilon if wt == 0 else wt
                wt = wt ** -1
        elif weight_options.get(weight_type) == weight_options.get('ANGLE'):
            """
            Edge angle centrality" in graph theory refers to a measure of an edge's importance within a network, 
            based on the angles formed between the edges connected to its endpoints, essentially assessing how "central" 
            an edge is in terms of its connection to other edges within the network, with edges forming more acute 
            angles generally considered more central. 
            To calculate edge angle centrality, you would typically:
               1. For each edge, identify the connected edges at its endpoints.
               2. Calculate the angles between these connected edges.
               3. Assign a higher centrality score to edges with smaller angles, indicating a more central position in the network structure.
            """
            sym_angle = np.minimum(pix_angle, (360 - pix_angle))
            wt = (sym_angle + epsilon) ** -1
        elif weight_options.get(weight_type) == weight_options.get('FIX_CON') or weight_options.get(weight_type) == weight_options.get('VAR_CON') or weight_options.get(weight_type) == weight_options.get('RES'):
            # Varies with width
            length = pix_length * pixel_dim
            area = math.pi * (pix_width * pixel_dim * 0.5) ** 2
            if weight_options.get(weight_type) == weight_options.get('FIX_CON'):
                area = math.pi * (1 * pixel_dim) ** 2
            num = length * rho_dim
            area = area + epsilon if area == 0 else area
            num =  num + epsilon if num == 0 else num
            wt = (num / area)  # Resistance
            if weight_options.get(weight_type) == weight_options.get('VAR_CON') or weight_options.get(weight_type) == weight_options.get('FIX_CON'):
                wt = wt ** -1  # Conductance is inverse of resistance
        else:
            raise TypeError('Invalid weight type')
        return pix_width, pix_angle, wt

    def _estimate_edge_width(self, graph_edge_coords: MatLike):
        """Estimates the edge width of a graph edge."""

        def find_orthogonal(u, v):
            # Inputs:
            # u, v: two coordinates (x, y) or (x, y, z)
            vec = u - v  # find the vector between u and v

            if np.linalg.norm(vec) == 0:
                n = np.array([0, ] * len(u), dtype=np.float16)
            else:
                # make n a unit vector along u,v
                n = vec / np.linalg.norm(vec)

            hl = np.linalg.norm(vec) / 2  # find the half-length of the vector u,v
            ortho_arr = np.random.randn(len(u))  # take a random vector
            ortho_arr -= ortho_arr.dot(n) * n  # make it orthogonal to vector u,v
            ortho_arr /= np.linalg.norm(ortho_arr)  # make it a unit vector

            # Returns the coordinates of the vector u,v midpoint; the orthogonal unit vector
            return (v + n * hl), ortho_arr

            # 1. Estimate orthogonal and mid-point
        end_index = len(graph_edge_coords) - 1
        pt1 = graph_edge_coords[0]
        pt2 = graph_edge_coords[end_index]
        # mid_index = int(len(graph_edge_coords) / 2)
        # mid_pt = graph_edge_coords[mid_index]

        mid_pt, ortho = find_orthogonal(pt1, pt2)
        # mid: the midpoint of a trace of an edge
        # ortho: an orthogonal unit vector
        mid_pt = mid_pt.astype(int)

        # 2. Compute the angle in Radians
        # Delta X and Y: Compute the difference in x and y coordinates:
        dx = pt2[0] - pt1[0]
        dy = pt2[1] - pt1[1]
        # Angle Calculation: Use the arc-tangent function to get the angle in radians:
        angle_rad = math.atan2(dy, dx)
        angle_deg = math.degrees(angle_rad)
        if angle_deg < 0:
            angle_deg += 360

        # 3. Estimate width
        check = 0  # initializing boolean check
        i = 0      # initializing iterative variable
        l1 = np.nan
        l2 = np.nan
        while check == 0:             # iteratively check along orthogonal vector to see if the coordinate is either...
            pt_check = mid_pt + i * ortho  # ... out of bounds, or no longer within the fiber in img_bin
            pt_check = pt_check.astype(int)
            is_in_edge = GraphSkeleton.point_check(self._img_bin, pt_check)

            if is_in_edge:
                edge = mid_pt + (i - 1) * ortho
                edge = edge.astype(int)
                l1 = edge  # When the check indicates oob or black space, assign width to l1
                check = 1
            else:
                i += 1

        check = 0
        i = 0
        while check == 0:  # Repeat, but following the negative orthogonal vector
            pt_check = mid_pt - i * ortho
            pt_check = pt_check.astype(int)
            is_in_edge = GraphSkeleton.point_check(self._img_bin, pt_check)

            if is_in_edge:
                edge = mid_pt - (i - 1) * ortho
                edge = edge.astype(int)
                l2 = edge  # When the check indicates oob or black space, assign width to l2
                check = 1
            else:
                i += 1

        # returns the length between l1 and l2, which is the width of the fiber associated with an edge, at its midpoint
        edge_width = np.linalg.norm(l1 - l2)
        return edge_width, angle_deg

    @classmethod
    def _generate_transformations(cls, pattern):
        """
        Generate common transformations for a pattern.

         * flipud is flipping them up-down
         * t_branch_2 is t_branch_0 transposed, which permutes it in all directions (might not be using that word right)
         * t_branch_3 is t_branch_2 flipped left right
         * those 3 functions are used to create all possible branches with just a few starting arrays below

        :param pattern: Pattern of the box as a numpy array.

        """
        return [
            pattern,
            np.flipud(pattern),
            np.fliplr(pattern),
            np.fliplr(np.flipud(pattern)),
            pattern.T,
            np.flipud(pattern.T),
            np.fliplr(pattern.T),
            np.fliplr(np.flipud(pattern.T))
        ]

    @classmethod
    def get_branched_points(cls, skeleton: MatLike):
        """Identify and retrieve the branched points from the graph skeleton."""
        skel_int = skeleton * 1

        # Define base patterns
        base_patterns = [
            [[1, 0, 1], [0, 1, 0], [1, 0, 1]],  # x_branch
            [[0, 1, 0], [1, 1, 1], [0, 1, 0]],  # x_branch variant
            [[0, 0, 0], [1, 1, 1], [0, 1, 0]],  # t_branch
            [[1, 0, 1], [0, 1, 0], [1, 0, 0]],  # t_branch variant
            [[1, 0, 1], [0, 1, 0], [0, 1, 0]],  # y_branch
            [[0, 1, 0], [1, 1, 0], [0, 0, 1]],  # y_branch variant
            [[0, 1, 0], [1, 1, 0], [1, 0, 1]],  # off_branch
            [[0, 1, 1], [0, 1, 1], [1, 0, 0]],  # clust_branch
            [[1, 1, 1], [0, 1, 1], [1, 0, 0]],  # clust_branch variant
            [[1, 1, 1], [0, 1, 1], [1, 0, 1]],  # clust_branch variant
            [[1, 0, 0], [1, 1, 1], [0, 1, 0]]  # cross_branch
        ]

        # Generate all transformations
        all_patterns = []
        for pattern in base_patterns:
            all_patterns.extend(cls._generate_transformations(np.array(pattern)))

        # Remove duplicate patterns (if any)
        unique_patterns = []
        for pattern in all_patterns:
            if not any(np.array_equal(pattern, existing) for existing in unique_patterns):
                unique_patterns.append(pattern)

        # Apply binary hit-or-miss for all unique patterns
        br = sum(ndimage.binary_hit_or_miss(skel_int, pattern) for pattern in unique_patterns)
        return br

    @classmethod
    def get_end_points(cls, skeleton: MatLike):
        """
        Identify and retrieve the end points from the graph skeleton.
        """
        skel_int = skeleton * 1

        # List of endpoint patterns
        endpoints = [
            [[0, 0, 0], [0, 1, 0], [0, 1, 0]],
            [[0, 0, 0], [0, 1, 0], [0, 0, 1]],
            [[0, 0, 0], [0, 1, 1], [0, 0, 0]],
            [[0, 0, 1], [0, 1, 0], [0, 0, 0]],
            [[0, 1, 0], [0, 1, 0], [0, 0, 0]],
            [[1, 0, 0], [0, 1, 0], [0, 0, 0]],
            [[0, 0, 0], [1, 1, 0], [0, 0, 0]],
            [[0, 0, 0], [0, 1, 0], [1, 0, 0]],
            [[0, 0, 0], [0, 1, 0], [0, 0, 0]]
        ]

        # Apply binary hit-or-miss for each pattern and sum results
        ep = sum(ndimage.binary_hit_or_miss(skel_int, np.array(pattern)) for pattern in endpoints)
        return ep

    @classmethod
    def prune_edges(cls, skeleton: MatLike, max_num, branch_points):
        """Prune dangling edges around b_points. Remove iteratively end points 'size' times from the skeleton"""
        temp_skeleton = skeleton.copy()
        for i in range(0, max_num):
            end_points = GraphSkeleton.get_end_points(temp_skeleton)
            points = np.logical_and(end_points, branch_points)
            end_points = np.logical_xor(end_points, points)
            end_points = np.logical_not(end_points)
            temp_skeleton = np.logical_and(temp_skeleton, end_points)
        return temp_skeleton

    @classmethod
    def merge_nodes(cls, skeleton: MatLike, node_radius):
        """Merge nearby nodes in the graph skeleton."""
        # overlay a disk over each branch point and find the overlaps to combine nodes
        skeleton_int = 1 * skeleton
        mask_elem = disk(node_radius)
        bp_skel = GraphSkeleton.get_branched_points(skeleton)
        bp_skel = 1 * (dilate(bp_skel, mask_elem))

        # wide-nodes is initially an empty image the same size as the skeleton image
        skel_shape = skeleton_int.shape
        wide_nodes = np.zeros(skel_shape, dtype='int')

        # this overlays the two skeletons
        # skeleton_integer is the full map, bp_skel is just the branch points blown up to a larger size
        for x in range(skel_shape[0]):
            for y in range(skel_shape[1]):
                if skeleton_int[x, y] == 0 and bp_skel[x, y] == 0:
                    wide_nodes[x, y] = 0
                else:
                    wide_nodes[x, y] = 1

        # re-skeletonizing wide-nodes and returning it, nearby nodes in radius 2 of each other should have been merged
        temp_skeleton = skeletonize(wide_nodes)
        return temp_skeleton

    @classmethod
    def remove_bubbles(cls, img_bin: MatLike, mask_elements: list):
        """
        Remove bubbles from the graph skeleton.
        Acknowledgements: Alain Kadar (https://github.com/compass-stc/StructuralGT/)
        """

        canvas = img_bin.copy()
        for mask_elem in mask_elements:
            canvas = skeletonize(mask_elem)
            canvas = binary_closing(canvas, footprint=mask_elem)

        temp_skeleton = skeletonize(canvas)
        return temp_skeleton

    @staticmethod
    def point_check(img_bin: MatLike, pt_check):
        """Checks and verifies that a point is on a graph edge."""

        def boundary_check(coord, w, h, d=None):
            """

            Args:
                coord: the coordinate (x,y) to check; no (x,y,z) compatibility yet.
                w: width of the image to set the boundaries.
                h: the height of the image to set the boundaries.
                d: the depth of the image to set the boundaries.
            Returns:

            """
            out_of_bounds = 0  # Generate a boolean check for out-of-boundary
            # Check if coordinate is within the boundary
            if d is None:
                if coord[0] < 0 or coord[1] < 0 or coord[-2] > (w - 1) or coord[-1] > (h - 1):
                    out_of_bounds = 1
            else:
                # if sum(coord < 0) > 0 or sum(coord > [w - 1, h - 1, d - 1]) > 0:
                if sum(coord < 0) > 0 or coord[-3] > (d - 1) or coord[-2] > (w - 1) or coord[-1] > (h - 1):
                    out_of_bounds = 1

            # returns the boolean oob (1 if there is a boundary error); coordinates (resets to (1,1) if boundary error)
            return out_of_bounds

        # Check if the image is 2D
        if len(img_bin.shape) == 2:
            is_2d = True
            height, width = img_bin.shape  # finds dimensions of img_bin for boundary check
            depth = 0
        else:
            is_2d = False
            depth, height, width = img_bin.shape

        try:
            if is_2d:
                # Checks if the point in fiber is out-of-bounds (oob) or black space (img_bin(x,y) = 0)
                oob = boundary_check(pt_check, width, height)
                not_in_edge = True if (oob == 1) else True if (img_bin[pt_check[-2], pt_check[-1]] == 0) else False
            else:
                # Checks if the point in fiber is out-of-bounds (oob) or black space (img_bin(d,x,y) = 0)
                oob = boundary_check(pt_check, width, height, d=depth)
                not_in_edge = True if (oob == 1) else True if (img_bin[pt_check[-3], pt_check[-2], pt_check[-1]] == 0) else False
        except IndexError:
            not_in_edge = True
        return not_in_edge



# SPDX-License-Identifier: GNU GPL v3

"""
Uses listener functions to send updates to outside functions.
"""


class ProgressUpdate:
    """
    A class for sending updates to outside functions.

    """
    def __init__(self):
        """
        A class for sending updates to outside functions.

        >>> def print_progress(code, msg):
        >>>     print(str(code) + ': ' + str(msg))
        >>>
        >>> upd = ProgressUpdate()
        >>> upd.add_listener(print_progress)  # to get updates
        >>> upd.update_status([1,"Sending update ..."])
        >>> upd.remove_listener(print_progress) # to opt out of updates
        >>>
        """
        self.__listeners = []
        self.abort = False

    def abort_tasks(self) -> None:
        """
        Set abort flag.
        :return:
        """
        self.abort = True

    def add_listener(self, func) -> None:
        """
        Add functions from the list of listeners.
        :param func:
        :return:
        """
        if func in self.__listeners:
            return
        self.__listeners.append(func)

    def remove_listener(self, func) -> None:
        """
        Remove functions from the list of listeners.
        :param func:
        :return:
        """
        if func not in self.__listeners:
            return
        self.__listeners.remove(func)

    def update_status(self, args=None) -> None:
        """
        Run all the functions that are saved as listeners.

        :param args:
        :return:
        """
        # Trigger events.
        if args is None:
            args = []
        for func in self.__listeners:
            func(*args)



"""

This code is a modified version of the original `sknw.py` by Yan Xiaolong.

Modifications were made to address bugs caused by 'numba' on some platforms (e.g., Windows) and to replace fixed-size
NumPy buffers with dynamically resizable ones, improving robustness for large images and graphs.

Original author: Yan Xiaolong
Original source: https://github.com/Image-Py/sknw.git

"""


import numpy as np
# from numba import jit
import networkx as nx


def neighbors(shape):
    dim = len(shape)
    block = np.ones([3] * dim)
    block[tuple([1] * dim)] = 0
    idx = np.where(block > 0)
    idx = np.array(idx, dtype=np.uint8).T
    idx = np.array(idx - [1] * dim)
    acc = np.cumprod((1,) + shape[::-1][:-1])
    return np.dot(idx, acc[::-1])


# @jit(nopython=True)  # my mark
def mark(img, nbs):  # mark the array use (0, 1, 2)
    img = img.ravel()
    for p in range(len(img)):
        if img[p] == 0: continue
        s = 0
        for dp in nbs:
            if img[p + dp] != 0: s += 1
        if s == 2:
            img[p] = 1
        else:
            img[p] = 2


# @jit(nopython=True)  # trans index to r, c...
def idx2rc(idx, acc):
    rst = np.zeros((len(idx), len(acc)), dtype=np.int16)
    for i in range(len(idx)):
        for j in range(len(acc)):
            rst[i, j] = idx[i] // acc[j]
            idx[i] -= rst[i, j] * acc[j]
    rst -= 1
    return rst


# @jit(nopython=True)  # fill a node (maybe two or more points)
def fill(img, p, num, nbs, acc):
    cap = 131072
    buf = np.empty(cap, dtype=np.int64)
    img[p] = num
    buf[0] = p
    cur = 0
    s = 1
    iso = True

    while cur < s:
        p = buf[cur]
        for dp in nbs:
            cp = p + dp
            if img[cp] == 2:
                img[cp] = num
                if s >= cap:
                    cap *= 2
                    buf = np.resize(buf, cap)
                buf[s] = cp
                s += 1
            elif img[cp] == 1:
                iso = False
        cur += 1

    return iso, idx2rc(buf[:s], acc)


# @jit(nopython=True)  # trace the edge and use a buffer, then buf.copy if you use [] numba not works
def trace(img, p, nbs, acc):
    cap = 2048
    buf = np.empty(cap, dtype=np.int64)
    cur = 1
    c1 = 0
    c2 = 0
    newp = 0

    while True:
        if cur >= cap:
            cap *= 2
            buf = np.resize(buf, cap)
        buf[cur] = p
        img[p] = 0
        cur += 1

        for dp in nbs:
            cp = p + dp
            if img[cp] >= 10:
                if c1 == 0:
                    c1 = img[cp]
                    buf[0] = cp
                else:
                    c2 = img[cp]
                    if cur >= cap:
                        cap += 1
                        buf = np.resize(buf, cap)
                    buf[cur] = cp
            elif img[cp] == 1:
                newp = cp

        p = newp
        if c2 != 0:
            break

    return c1 - 10, c2 - 10, idx2rc(buf[:cur + 1], acc)


# @jit(nopython=True)  # parse the image, then get the nodes and edges
def parse_struc(img, nbs, acc, iso, ring):
    img = img.ravel()
    # buf = np.zeros(13107200, dtype=np.int64)
    num = 10
    nodes = []
    for p in range(len(img)):
        if img[p] == 2:
            isiso, nds = fill(img, p, num, nbs, acc)
            if isiso and not iso: continue
            num += 1
            nodes.append(nds)
    edges = []
    for p in range(len(img)):
        if img[p] < 10: continue
        for dp in nbs:
            if img[p + dp] == 1:
                edge = trace(img, p + dp, nbs, acc)
                edges.append(edge)
    if not ring: return nodes, edges
    for p in range(len(img)):
        if img[p] != 1: continue
        img[p] = num
        num += 1
        nodes.append(idx2rc([p], acc))
        for dp in nbs:
            if img[p + dp] == 1:
                edge = trace(img, p + dp, nbs, acc)
                edges.append(edge)
    return nodes, edges


# use nodes and edges to build a networkx graph
def build_graph(nodes, edges, multi=False, full=True):
    os = np.array([i.mean(axis=0) for i in nodes])
    if full: os = os.round().astype(np.uint16)
    graph = nx.MultiGraph() if multi else nx.Graph()
    for i in range(len(nodes)):
        graph.add_node(i, pts=nodes[i], o=os[i])
    for s, e, pts in edges:
        if full: pts[[0, -1]] = os[[s, e]]
        l = np.linalg.norm(pts[1:] - pts[:-1], axis=1).sum()
        graph.add_edge(s, e, pts=pts, weight=l)
    return graph


def mark_node(ske):
    buf = np.pad(ske, (1, 1), mode='constant').astype(np.uint16)
    nbs = neighbors(buf.shape)
    # acc = np.cumprod((1,) + buf.shape[::-1][:-1])[::-1]
    mark(buf, nbs)
    return buf


def build_sknw(ske, multi=False, iso=True, ring=True, full=True):
    buf = np.pad(ske, (1, 1), mode='constant').astype(np.uint16)
    nbs = neighbors(buf.shape)
    acc = np.cumprod((1,) + buf.shape[::-1][:-1])[::-1]
    mark(buf, nbs)
    nodes, edges = parse_struc(buf, nbs, acc, iso, ring)
    return build_graph(nodes, edges, multi, full)


# draw the graph
def draw_graph(img, graph, cn=255, ce=128):
    acc = np.cumprod((1,) + img.shape[::-1][:-1])[::-1]
    img = img.ravel()
    for (s, e) in graph.edges():
        eds = graph[s][e]
        if isinstance(graph, nx.MultiGraph):
            for i in eds:
                pts = eds[i]['pts']
                img[np.dot(pts, acc)] = ce
        else:
            img[np.dot(eds['pts'], acc)] = ce
    for idx in graph.nodes():
        pts = graph.nodes[idx]['pts']
        img[np.dot(pts, acc)] = cn



# SPDX-License-Identifier: GNU GPL v3

"""
A group of algorithms and functions for Graph Theory analysis on microscopy images.
"""

# MODULES
from .apps.lib_app import ExpressGT
from .compute.graph_analyzer import GraphAnalyzer
from .imaging.base_image import BaseImage
from .imaging.image_processor import (
    ImageProcessor,
    ALLOWED_IMG_EXTENSIONS,
    ALLOWED_GRAPH_FILE_EXTENSIONS
)
from .models.filter_env import FilterSearchSpace
from .models.mdp_env import SGTGraphEnv
from .models.filter_env import (
    sgt_genetic_algorithm,
    sgt_hill_climbing_algorithm
)
from .networks.fiber_network import FiberNetworkBuilder
from .networks.graph_skeleton import GraphSkeleton
from .utils.sgt_utils import (
    gsd_to_skeleton,
    csv_to_graph,
    write_gsd_file
)
from .utils.config_loader import (
    load_gtc_configs,
    load_gte_configs,
    load_img_configs
)

__all__ = [
    "ExpressGT",
    "BaseImage",
    "GraphAnalyzer",
    "ImageProcessor",
    "ALLOWED_IMG_EXTENSIONS",
    "ALLOWED_GRAPH_FILE_EXTENSIONS",
    "FiberNetworkBuilder",
    "GraphSkeleton",
    "FilterSearchSpace",
    "SGTGraphEnv",
    "load_gtc_configs",
    "load_gte_configs",
    "load_img_configs",
    "gsd_to_skeleton",
    "csv_to_graph",
    "write_gsd_file",
    "sgt_genetic_algorithm",
    "sgt_hill_climbing_algorithm"
]



# SPDX-License-Identifier: GNU GPL v3

"""
Base worker class for executing all resource-intensive StructuralGT tasks.
"""

import logging
from PySide6.QtCore import QObject, Signal
from ..compute.graph_analyzer import GraphAnalyzer
from ..utils.sgt_utils import AbortException, plot_to_opencv


class BaseWorker(QObject):

    inProgressSignal = Signal(int, str)         # progress-value (0-100), progress-message (str)
    taskFinishedSignal = Signal(bool, object)    # success/fail (True/False), result (object)

    def __init__(self):
        super().__init__()

    def update_progress(self, value, msg):
        """
        Send the update_progress signal to all listeners.
        Progress-value (0-100), progress-message (str)
        Args:
            value: progress value (0-100), (-1, if it is an error), (101, if it is the nav-control message)
            msg: progress message (str)

        Returns:

        """
        self.inProgressSignal.emit(value, msg)

    def task_save_images(self, ntwk_p):
        """"""
        try:
            self.update_progress(25, "Saving Images...")
            ntwk_p.save_images_to_file()
            self.update_progress(95, "Saving Images...")
            self.taskFinishedSignal.emit(True, "Image files successfully saved in 'Output Dir'")
        except Exception as err:
            logging.exception("Error: %s", err, extra={'user': 'SGT Logs'})
            self.taskFinishedSignal.emit(False, ["Save Images Failed", "Error while saving images!"])

    def task_export_graph(self, ntwk_p):
        """"""
        try:
            # 1. Get filename
            self.update_progress(25, "Exporting Graph...")
            filename, out_dir = ntwk_p.get_filenames()

            # 2. Save graph data to the file
            self.update_progress(30, "Exporting Graph...")
            ntwk_p.graph_obj.save_graph_to_file(filename, out_dir)
            self.update_progress(95, "Exporting Graph...")
            self.taskFinishedSignal.emit(True, "Graph successfully exported to file and saved in 'Output Dir'")
        except Exception as err:
            logging.exception("Error: %s", err, extra={'user': 'SGT Logs'})
            self.taskFinishedSignal.emit(False, ["Export Graph Failed", "Error while exporting graph!"])

    def task_apply_img_filters(self, ntwk_p):
        """"""
        try:
            ntwk_p.add_listener(self.update_progress)
            ntwk_p.apply_img_filters()
            ntwk_p.remove_listener(self.update_progress)
            self.taskFinishedSignal.emit(True, ntwk_p)
        except Exception as err:
            logging.exception("Error: %s", err, extra={'user': 'SGT Logs'})
            # self.abort = True
            self.update_progress(-1, "Error encountered! Try again")
            # Emit failure signal (aborted)
            self.taskFinishedSignal.emit(False, ["Apply Filters Failed", "Fatal error while applying filters! "
                                                                         "Change filter settings and try again; "
                                                                         "Or, Close the app and try again."])

    def task_calculate_img_histogram(self, ntwk_p):
        """"""
        try:
            hist_images = [plot_to_opencv(obj.plot_img_histogram(curr_view=ntwk_p.selected_batch_view)) for obj in ntwk_p.image_obj_3d]
            self.taskFinishedSignal.emit(True, hist_images)
        except Exception as err:
            logging.exception("Error: %s", err, extra={'user': 'SGT Logs'})
            self.taskFinishedSignal.emit(False, ["Histogram Calculation Failed", "Error while calculating image histogram!"])

    def task_extract_graph(self, ntwk_p):
        """"""
        try:
            ntwk_p.abort = False
            ntwk_p.add_listener(self.update_progress)
            ntwk_p.apply_img_filters()
            ntwk_p.build_graph_network()
            if ntwk_p.abort:
                raise AbortException("Process aborted")
            ntwk_p.remove_listener(self.update_progress)
            self.taskFinishedSignal.emit(True, ntwk_p)
        except AbortException as err:
            logging.exception("Task Aborted: %s", err, extra={'user': 'SGT Logs'})
            # Clean up listeners before exiting
            ntwk_p.remove_listener(self.update_progress)
            # Emit failure signal (aborted)
            self.taskFinishedSignal.emit(False, ["Extract Graph Aborted", "Graph extraction aborted due to error! "
                                                                          "Change image filters and/or graph settings "
                                                                          "and try again. If error persists then close "
                                                                          "the app and try again."])
        except Exception as err:
            logging.exception("Error: %s", err, extra={'user': 'SGT Logs'})
            self.update_progress(-1, "Error encountered! Try again")
            # Clean up listeners before exiting
            ntwk_p.remove_listener(self.update_progress)
            # Emit failure signal (aborted)
            self.taskFinishedSignal.emit(False, ["Extract Graph Failed", "Graph extraction aborted due to error! "
                                                                          "Change image filters and/or graph settings "
                                                                          "and try again. If error persists then close "
                                                                          "the app and try again."])

    def task_compute_gt(self, sgt_obj):
        """"""
        success, new_sgt = GraphAnalyzer.safe_run_analyzer(sgt_obj, self.update_progress, save_to_pdf=True)
        if success:
            self.taskFinishedSignal.emit(True, new_sgt)
        else:
            self.taskFinishedSignal.emit(False, ["SGT Computations Failed", "Fatal error occurred while computing GT parameters. Change image filters and/or graph settings and try again. If error persists then close the app and try again."])

    def task_compute_multi_gt(self, sgt_objs):
        """"""
        new_sgt_objs = GraphAnalyzer.safe_run_multi_analyzer(sgt_objs, self.update_progress)
        if new_sgt_objs is not None:
            self.taskFinishedSignal.emit(True, sgt_objs)
        else:
            msg = "Either task was aborted by user or a fatal error occurred while computing GT parameters. Change image filters and/or graph settings and try again. If error persists then close the app and try again."
            self.taskFinishedSignal.emit(False, ["SGT Computations Aborted/Failed", msg])


import platform
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext


class BuildExt(build_ext):
    """
        Custom build_ext to handle different OS configurations.
    """

    def build_extensions(self):
        # Use MinGW (Windows) or system compiler (Linux/macOS)
        if platform.system() == "Windows":
            print("Configuring build for Windows (MinGW)...")
            #for ext in self.extensions:
            #    ext.extra_compile_args = ["-Wall", "-O2"]
            #    ext.extra_link_args = ["-static"]
        elif platform.system() == "Darwin":
            print("Configuring build for macOS...")
            for ext in self.extensions:
                ext.extra_compile_args = ["-std=c99", "-O2"]
        elif platform.system() == "Linux":
            print("Configuring build for Linux...")
            # Make sure: export CC=/home/linuxbrew/.linuxbrew/bin/gcc-15
            # Make sure to add in PyCharm Run Environment: LD_LIBRARY_PATH=/home/linuxbrew/.linuxbrew/lib:$LD_LIBRARY_PATH
            for ext in self.extensions:
                ext.extra_compile_args = ["-std=c99", "-O2"]
        build_ext.build_extensions(self)


# Make sure igraph lib is installed
# brew install igraph - macOS (Homebrew)
# brew install igraph - Linux (Homebrew)
# Build with CMAKE    - Windows
ext_modules = [
    Extension(
        name="sgtlib.compute.c_lang.sgt_c_module",
        sources=["src/sgtlib/compute/c_lang/sgtmodule.c", "src/sgtlib/compute/c_lang/sgt_base.c"],
        libraries=["igraph"],  # macOS/Linux
        include_dirs=["src/sgtlib/compute/c_lang/include"],
    )
]

# Windows-specific settings (make sure MinGW is installed in location "C:MinGW")
if platform.system() == "Windows":
    ext_modules[0].libraries = ["igraph-x64", "libpthreadVCE3-x64"]
    ext_modules[0].include_dirs = ["C:/msys64/ucrt64/include/igraph", "C:/msys64/ucrt64/include/pthread"]
    ext_modules[0].library_dirs = ["C:/msys64/ucrt64/lib/igraph", "C:/msys64/ucrt64/lib/pthread"]
    ext_modules[0].extra_link_args = ["/VERBOSE:LIB"]

if platform.system() == "Darwin":
    ext_modules[0].include_dirs = ["/opt/homebrew/Cellar/igraph/0.10.16/include/igraph"]
    ext_modules[0].library_dirs = ["/opt/homebrew/Cellar/igraph/0.10.16/lib"]

if platform.system() == "Linux":
    ext_modules[0].include_dirs = ["/home/linuxbrew/.linuxbrew/Cellar/igraph/0.10.15_1/include/igraph"]
    ext_modules[0].library_dirs = ["/home/linuxbrew/.linuxbrew/Cellar/igraph/0.10.15_1/lib"]


# Setup configuration
setup(
    #ext_modules=ext_modules,
    #cmdclass={"build_ext": BuildExt},  # Use the custom build class
    # **extra_options
)



# SPDX-License-Identifier: GNU GPL v3
"""
A class for building an MDP environment of image filters for graph generation.
"""



class SGTGraphEnv:
    """
    Class for building a (Markov Decision Process) MDP environment for manipulating the Genetic Algorithm (GA) optimizer
    into navigating the search space of image filters.
    """

    def __init__(self):
        pass

    def _build_action_space(self):
        pass

    def _build_observation_space(self):
        pass



# SPDX-License-Identifier: GNU GPL v3

"""
StructuralGT utility functions.
"""

import os
import io
import sys
import cv2
import base64
import random
# import socket
import logging
# import platform
import subprocess
import gsd.hoomd
import numpy as np
import pandas as pd
import networkx as nx
import multiprocessing as mp
import matplotlib.pyplot as plt
from PIL import Image
from typing import LiteralString
from cv2.typing import MatLike


class AbortException(Exception):
    """Custom exception to handle task cancellation initiated by the user or an error."""
    pass


def get_num_cores() -> int | bool:
    """
    Finds the count of CPU cores in a computer or a SLURM supercomputer.
    :return: Number of cpu cores (int)
    """

    def __get_slurm_cores__():
        """
        Test the computer to see if it is a SLURM environment, then gets the number of CPU cores.
        :return: Count of CPUs (int) or False
        """
        try:
            cores = int(os.environ['SLURM_JOB_CPUS_PER_NODE'])
            return cores
        except ValueError:
            try:
                str_cores = str(os.environ['SLURM_JOB_CPUS_PER_NODE'])
                temp = str_cores.split('(', 1)
                cpus = int(temp[0])
                str_nodes = temp[1]
                temp = str_nodes.split('x', 1)
                str_temp = str(temp[1]).split(')', 1)
                nodes = int(str_temp[0])
                cores = cpus * nodes
                return cores
            except ValueError:
                return False
        except KeyError:
            return False

    num_cores = __get_slurm_cores__()
    if not num_cores:
        num_cores = mp.cpu_count()
    return int(num_cores)


def verify_path(a_path) -> tuple[bool, str]:
    if not a_path:
        return False, "No folder/file selected."

    # Convert QML "file:///" path format to a proper OS path
    if a_path.startswith("file:///"):
        if sys.platform.startswith("win"):
            # Windows Fix (remove extra '/')
            a_path = a_path[8:]
        else:
            # macOS/Linux (remove "file://")
            a_path = a_path[7:]

    # Normalize the path
    a_path = os.path.normpath(a_path)

    if not os.path.exists(a_path):
        return False, f"File/Folder in {a_path} does not exist. Try again."
    return True, a_path


def install_package(package) -> None:
    try:
        subprocess.check_call([sys.executable, "-m", "pip", "install", package])
        logging.info(f"Successfully installed {package}", extra={'user': 'SGT Logs'})
    except subprocess.CalledProcessError:
        logging.info(f"Failed to install {package}: ", extra={'user': 'SGT Logs'})


def detect_cuda_version() -> str | None:
    """Check if CUDA is installed and return its version."""
    try:
        output = subprocess.check_output(['nvcc', '--version']).decode()
        if 'release 12' in output:
            return '12'
        elif 'release 11' in output:
            return '11'
        else:
            return None
    except (subprocess.CalledProcessError, FileNotFoundError):
        logging.info(f"Please install 'NVIDIA GPU Computing Toolkit' via: https://developer.nvidia.com/cuda-downloads", extra={'user': 'SGT Logs'})
        return None


"""
def detect_cuda_and_install_cupy():
    try:
        import cupy
        logging.info(f"CuPy is already installed: {cupy.__version__}", extra={'user': 'SGT Logs'})
        return
    except ImportError:
        logging.info("CuPy is not installed.", extra={'user': 'SGT Logs'})

    def is_connected(host="8.8.8.8", port=53, timeout=3):
        # Check if the system has an active internet connection.
        try:
            socket.setdefaulttimeout(timeout)
            socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
            return True
        except socket.error:
            return False

    if not is_connected():
        logging.info("No internet connection. Cannot install CuPy.", extra={'user': 'SGT Logs'})
        return

    # Handle macOS (Apple Silicon) - CPU only
    if platform.system() == "Darwin" and platform.processor().startswith("arm"):
        logging.info("Detected MacOS with Apple Silicon (M1/M2/M3). Installing CPU-only version of CuPy.", extra={'user': 'SGT Logs'})
        # install_package('cupy')  # CPU-only version
        return

    # Handle CUDA systems (Linux/Windows with GPU)
    cuda_version = detect_cuda_version()

    if cuda_version:
        logging.info(f"CUDA detected: {cuda_version}", extra={'user': 'SGT Logs'})
        if cuda_version == '12':
            install_package('cupy-cuda12x')
        elif cuda_version == '11':
            install_package('cupy-cuda11x')
        else:
            logging.info("CUDA version not supported. Installing CPU-only CuPy.", extra={'user': 'SGT Logs'})
            install_package('cupy')
    else:
        # No CUDA found, fall back to the CPU-only version
        logging.info("CUDA not found. Installing CPU-only CuPy.", extra={'user': 'SGT Logs'})
        install_package('cupy')

    # Proceed with installation if connected
    cuda_version = detect_cuda_version()
    if cuda_version == '12':
        install_package('cupy-cuda12x')
    elif cuda_version == '11':
        install_package('cupy-cuda11x')
    else:
        logging.info("No CUDA detected or NVIDIA GPU Toolkit not installed. Installing CPU-only CuPy.", extra={'user': 'SGT Logs'})
        install_package('cupy')
"""


def write_txt_file(data: str, path: LiteralString | str | bytes, wr=True) -> None:
    """Description
        Writes data into a txt file.

        :param data: Information to be written
        :param path: name of the file and storage path
        :param wr: writes data into file if True
        :return:
    """
    if wr:
        with open(path, 'w') as f:
            f.write(data)
            f.close()
    else:
        pass


def write_gsd_file(f_name: str, skeleton: np.ndarray) -> None:
    """
    A function that writes graph particles to a GSD file. Visualize with OVITO software.
    Acknowledgements: Alain Kadar (https://github.com/compass-stc/StructuralGT/)

    :param f_name: gsd.hoomd file name
    :param skeleton: skimage.morphology skeleton
    """
    # pos_count = int(sum(skeleton.ravel()))
    particle_positions = np.asarray(np.where(np.asarray(skeleton) != 0)).T
    with gsd.hoomd.open(name=f_name, mode="w") as f:
        s = gsd.hoomd.Frame()
        s.particles.N = len(particle_positions)  # OR pos_count
        s.particles.position = particle_positions
        s.particles.types = ["A"]
        s.particles.typeid = ["0"] * s.particles.N
        f.append(s)


def gsd_to_skeleton(gsd_file: str, is_2d:bool=False) -> None | np.ndarray:
    """
    A function that takes a gsd file and returns a NetworkX graph object.
    Acknowledgements: Alain Kadar (https://github.com/compass-stc/StructuralGT/)

    :param gsd_file: gsd.hoomd file name;
    :param is_2d: is the skeleton 2D?
    :return:
    """

    def shift(points: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """Translates all points such that the minimum coordinate in points is the origin.

        Args:
            points: The points to shift.

        Returns:
            The shifted points.
            The applied shift.
        """
        if is_2d:
            shifted_points = np.full(
                (np.shape(points)[0], 2),
                [np.min(points.T[0]), np.min(points.T[1])],
            )
        else:
            shifted_points = np.full(
                (np.shape(points)[0], 3),
                [
                    np.min(points.T[0]),
                    np.min(points.T[1]),
                    np.min(points.T[2]),
                ],
            )
        points = points - shifted_points
        return points, shifted_points

    def reduce_dim(all_positions: np.ndarray) -> np.ndarray:
        """For lists of positions where all elements along one axis have the same
        value, this returns the same list of positions but with the redundant
        dimension(s) removed.

        Args:
            all_positions: The positions to reduce.

        Returns:
            The reduced positions
        """

        unique_positions = np.asarray(
            list(len(np.unique(all_positions.T[i])) for i in range(len(all_positions.T)))
        )
        redundant = unique_positions == 1
        all_positions = all_positions.T[~redundant].T
        return all_positions

    frame = gsd.hoomd.open(name=gsd_file, mode="r")[0]
    positions = shift(frame.particles.position.astype(int))[0]

    if sum((positions < 0).ravel()) != 0:
        positions = shift(positions)[0]

    if is_2d:
        """
        is_2d (optional, bool):
            Whether the skeleton is 2D. If True it only ensures additional
            redundant axes from the position array is removed. It does not
            guarantee a 3d graph.
        """
        positions = reduce_dim(positions)
        new_pos = np.zeros(positions.T.shape)
        new_pos[0] = positions.T[0]
        new_pos[1] = positions.T[1]
        positions = new_pos.T.astype(int)

    skel_int = np.zeros(
        list((max(positions.T[i]) + 1) for i in list(
            range(min(positions.shape))))
    )
    skel_int[tuple(list(positions.T))] = 1
    return skel_int.astype(int)


def csv_to_graph(csv_path: str) -> None | nx.Graph:
    """
    Load a graph from a file that may contain:
      - Edge list (2 columns)
      - Adjacency matrix (square matrix)
      - XYZ positions (3 columns: x, y, z, edges inferred by distance threshold)

    :param csv_path: Path to the graph file
    """

    # Check if the first line is text (header) instead of numbers
    with open(csv_path, "r") as f:
        first_line = f.readline()
    try:
        [float(x) for x in first_line.replace(",", " ").split()]
        skip = 0  # numeric → no header
    except ValueError:
        skip = 1  # not numeric → skip header

    # Try to read as a numeric matrix
    try:
        data = np.loadtxt(csv_path, delimiter=",", dtype=np.float64, skiprows=skip)
    except ValueError:
        return None

    if data is None:
        return None

    # Case 1: Edge list (two columns)
    if data.ndim == 2 and data.shape[1] == 2:
        nx_graph = nx.Graph()
        for u, v in data.astype(int):
            nx_graph.add_edge(u, v)
        return nx_graph

    # Case 2: Adjacency matrix (square matrix)
    elif data.ndim == 2 and data.shape[0] == data.shape[1]:
        nx_graph = nx.from_numpy_array(data)
        return nx_graph

    # Case 3: XYZ positions (three columns)
    elif data.ndim == 2 and data.shape[1] == 3:
        from scipy.spatial import distance_matrix
        # Build graph based on proximity (set threshold distance)
        threshold = 1.0
        dist_mat = distance_matrix(data, data)
        nx_graph = nx.Graph()
        for i in range(len(data)):
            nx_graph.add_node(i, pos=data[i])
        for i in range(len(data)):
            for j in range(i + 1, len(data)):
                if dist_mat[i, j] < threshold:
                    nx_graph.add_edge(i, j, weight=dist_mat[i, j])
        return nx_graph
    else:
        return None


def img_to_base64(img: MatLike | Image.Image) -> MatLike | None:
    """ Converts a Numpy/OpenCV or PIL image to a base64 encoded string."""
    if img is None:
        return None

    if type(img) == np.ndarray:
        return opencv_to_base64(img)

    if type(img) == Image.Image:
        # Convert to numpy, apply safe conversion
        np_img = np.array(img)
        img_norm = safe_uint8_image(np_img)
        return opencv_to_base64(img_norm)
    return None


def opencv_to_base64(img_arr: MatLike) -> str | None:
    """Convert an OpenCV/Numpy image to a base64 string."""
    success, encoded_img = cv2.imencode('.png', img_arr)
    if success:
        buffer = io.BytesIO(encoded_img.tobytes())
        buffer.seek(0)
        base64_data = base64.b64encode(buffer.getvalue()).decode("utf-8")
        return base64_data
    else:
        return None


def plot_to_opencv(fig: plt.Figure) -> MatLike | None:
    """Convert a Matplotlib figure to an OpenCV BGR image (Numpy array), retaining colors."""
    if fig:
        # Save a figure to a buffer
        buf = io.BytesIO()
        fig.savefig(buf, format='png', bbox_inches='tight')
        buf.seek(0)

        # Convert buffer to NumPy array
        img_array = np.frombuffer(buf.getvalue(), dtype=np.uint8)
        buf.close()

        # Decode image including the alpha channel (if any)
        img_cv_rgba = cv2.imdecode(img_array, cv2.IMREAD_UNCHANGED)

        # Convert RGBA to RGB if needed
        if img_cv_rgba.shape[2] == 4:
            img_cv_rgb = cv2.cvtColor(img_cv_rgba, cv2.COLOR_RGBA2RGB)
        else:
            img_cv_rgb = img_cv_rgba

        # Convert RGB to BGR to match OpenCV color space
        img_cv_bgr = cv2.cvtColor(img_cv_rgb, cv2.COLOR_RGB2BGR)
        return img_cv_bgr
    return None


def safe_uint8_image(img: MatLike) -> MatLike | None:
    """
    Converts an image to uint8 safely:
        - If already uint8, returns as is.
        - If float or other type, normalizes to 0–255 and converts to uint8.
    """
    if img is None:
        return None

    if img.dtype == np.uint8:
        return img

    # Handle float or other types
    min_val = float(np.min(img))
    max_val = float(np.max(img))

    if min_val == max_val:
        # Avoid divide by zero; return constant grayscale
        return np.full(img.shape, 0 if min_val == 0 else 255, dtype=np.uint8)

    # Normalize to 0–255
    norm_img = ((img - min_val) / (max_val - min_val)) * 255.0
    return norm_img.astype(np.uint8)


def sgt_excel_to_dataframe(excel_dir_path: str, allowed_ext: str = ".xlsx") -> dict[str, pd.DataFrame] | None:
    """
    Read the Excel files (generated by StructuralGT) and save in one DataFrame.

    :param excel_dir_path: Path to the Excel files directory.
    :param allowed_ext: Allowed file extensions (default: ".xlsx").
    :return: A dictionary of DataFrames, where the key is the Excel file name (without extension) and the value is the DataFrame.
    """

    if excel_dir_path is None:
        return None

    files = os.listdir(excel_dir_path)
    files = sorted(files)
    rename_map = {
        "Nodes-Number of edge.": "Nodes-Edges",
        "Nodes-Number of edge. (Fitting)": "Nodes-Edges(Fit)",
        "Nodes-Average degree": "Nodes-Degree",
        "Nodes-Average degree (Fitting)": "Nodes-Degree(Fit)",
        "Nodes-Network diamet.": "Nodes-Diameter",
        "Nodes-Network diamet. (Fitting)": "Nodes-Diameter(Fit)",
        "Nodes-Graph density": "Nodes-Density",
        "Nodes-Graph density (Fitting)": "Nodes-Density(Fit)",
        "Nodes-Average betwee.": "Nodes-BC",
        "Nodes-Average betwee. (Fitting)": "Nodes-BC(Fit)",
        "Nodes-Average eigenv.": "Nodes-EC",
        "Nodes-Average eigenv. (Fitting)": "Nodes-EC(Fit)",
        "Nodes-Average closen.": "Nodes-CC",
        "Nodes-Average closen. (Fitting)": "Nodes-CC(Fit)",
        "Nodes-Assortativity .": "Nodes-ASC",
        "Nodes-Assortativity . (Fitting)": "Nodes-ASC(Fit)",
        "Nodes-Average cluste.": "Nodes-ACC",
        "Nodes-Average cluste. (Fitting)": "Nodes-ACC(Fit)",
        "Nodes-Global efficie.": "Nodes-GE",
        "Nodes-Global efficie. (Fitting)": "Nodes-GE(Fit)",
        "Nodes-Wiener Index": "Nodes-WI",
        "Nodes-Wiener Index (Fitting)": "Nodes-WI(Fit)",
    }

    all_sheets = {}
    for a_file in files:
        if a_file.endswith(allowed_ext):
            # Get the Excel file and load its contents
            file_path = os.path.join(excel_dir_path, a_file)
            file_sheets = pd.read_excel(file_path, sheet_name=None)

            # Append Excel data to one place
            for sheet_name, df in file_sheets.items():
                # Rename it if sheet_name exists in mapping
                new_name = rename_map.get(sheet_name, sheet_name)

                # Add Material column with file name (without extension)
                df = df.copy()
                df["Material"] = os.path.splitext(a_file)[0]

                if new_name not in all_sheets:
                    all_sheets[new_name] = []  # initialize list
                all_sheets[new_name].append(df)

    # Concatenate each list of DataFrames into one
    for sheet_name in all_sheets:
        all_sheets[sheet_name] = pd.concat(all_sheets[sheet_name], ignore_index=True)
    return all_sheets


def gen_spider_plot(df_sgt: pd.DataFrame, materials: list[str], parameters: list[str]) -> None|plt.Figure:
    """
    Creates a spider plot that compares the GT parameters for each material.
    Args:
        df_sgt: A dataframe containing the SGT results.
        materials: A list of material names to include in the plot.
        parameters: A list of GT parameters to be compared.

    Returns: None or a Matplotlib figure.
    """

    if df_sgt is None or materials is None or parameters is None:
        return None

    param_rename_map = {
        "Number of nodes": "Nodes",
        "Number of edges": "Edges",
        "Network diameter": "Diameter",
        "Average edge angle (degrees)": "Avg. E. Angle",
        "Median edge angle (degrees)": "Med. E. Angle",
        "Graph density": "GD",
        "Average degree": "AD",
        "Global efficiency": "GE",
        "Wiener Index": "WI",
        "Assortativity coefficient": "ASC",
        "Average clustering coefficient": "ACC",
        "Average betweenness centrality": "BC",
        "Average eigenvector centrality": "EC",
        "Average closeness centrality": "CC",
    }
    value_cols = ["value-1", "value-2", "value-3", "value-4"]

    # Rename Columns: apply replacements in the "Parameter" column
    if "parameter" in df_sgt.columns:
        df_sgt["parameter"] = df_sgt["parameter"].replace(param_rename_map)

    # Ensure the value columns exist
    if all(col in df_sgt.columns for col in value_cols):
        df_sgt["Avg."] = df_sgt[value_cols].astype(float).mean(axis=1)
        df_sgt["Std. Dev."] = df_sgt[value_cols].astype(float).std(axis=1)

    # Filter and pivot
    df_avg = df_sgt.pivot(index='Material', columns='parameter', values='Avg.')
    df_std = df_sgt.pivot(index='Material', columns='parameter', values='Std. Dev.')

    # Ensure consistent parameter order
    df_avg = df_avg[parameters]
    df_std = df_std[parameters]

    # Radar chart setup
    num_vars = len(parameters)
    angles = np.linspace(0, 2 * np.pi, num_vars, endpoint=False).tolist()
    angles_closed = angles + [angles[0]]  # close the loop without mutating the input list

    # Create the figure and axes
    fig, ax = plt.subplots(figsize=(8, 8), subplot_kw=dict(polar=True))

    # Plot each material
    for mat in materials:
        values = df_avg.loc[mat].tolist()
        values += [values[0]]  # close the loop

        errors = df_std.loc[mat].tolist()
        errors += [errors[0]]

        ax.plot(angles_closed, values, label=mat)
        ax.fill_between(angles_closed,
                        np.array(values) - np.array(errors),
                        np.array(values) + np.array(errors),
                        alpha=0.1)

    # Final touches
    ax.set_theta_offset(np.pi / 2)
    ax.set_theta_direction(-1)
    ax.set_thetagrids(np.degrees(angles), parameters)
    ax.set_title("Spider Plot with Std. Dev. Error Bands", fontsize=14)
    ax.legend(loc='upper right', bbox_to_anchor=(1.3, 1.1))
    plt.tight_layout()
    return fig



from PySide6.QtCore import Qt, QAbstractListModel
from ...utils.sgt_utils import img_to_base64


class ImageGridModel(QAbstractListModel):
    IdRole = Qt.ItemDataRole.UserRole + 1
    SelectedRole = Qt.ItemDataRole.UserRole + 20
    ImageRole = Qt.ItemDataRole.UserRole + 21

    def __init__(self, img_lst: list, selected_images: set, parent=None):
        super().__init__(parent)
        if len(img_lst) == 0:
            self._image_data = []
            return
        self._image_data = [{
            "id": i,
            "image": img_to_base64(img_lst[i]) if img_lst[i] is not None else "",
            "selected": 1 if i in selected_images else 0
        } for i in range(len(img_lst))]

    def rowCount(self, parent=None):
        return len(self._image_data)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid() or index.row() >= len(self._image_data):
            return None

        item = self._image_data[index.row()]
        if item["image"] is None:
            return None

        if role == self.IdRole:
            return item["id"]
        elif role == self.ImageRole:
            return item["image"]
        elif role == self.SelectedRole:
            return item["selected"]
        return ""

    def setData(self, index, value, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid() or index.row() >= len(self._image_data):
            return False

        if role == self.SelectedRole:
            self._image_data[index.row()]["selected"] = value
            self.dataChanged.emit(index, index, [role])
            return True
        if role == self.ImageRole:
            self._image_data[index.row()]["image"] = value
            self.dataChanged.emit(index, index, [role])
            return True
        return False

    def reset_data(self, new_data: list, selected_images: set):
        """ Resets the data to be displayed. """
        if new_data is None:
            return
        self.beginResetModel()

        self._image_data = [{
            "id": i,
            "image": img_to_base64(new_data[i]) if new_data[i] is not None else "",
            "selected": 1 if i in selected_images else 0
        } for i in range(len(new_data))]

        self.endResetModel()
        self.dataChanged.emit(self.index(0, 0), self.index(len(new_data), 0),
                              [self.IdRole, self.ImageRole, self.SelectedRole])

    def roleNames(self):
        return {
            self.IdRole: b"id",
            self.ImageRole: b"image",
            self.SelectedRole: b"selected",
        }
